;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Serielle Schnittstelle:

 
.equ SERIAL_UBRR_VAL   = ((F_CPU+SERIAL_BAUD*8)/(SERIAL_BAUD*16)-1)  ; clever runden
.equ SERIAL_BAUD_REAL  = (F_CPU/(16*(SERIAL_UBRR_VAL+1)))      ; Reale Baudrate
.equ SERIAL_BAUD_ERROR = ((SERIAL_BAUD_REAL*1000)/SERIAL_BAUD-1000)  ; Fehler in Promille
 
.if ((SERIAL_BAUD_ERROR>10) || (SERIAL_BAUD_ERROR<-10))       ; max. +/-10 Promille Fehler
  .error "Systematischer Fehler der Baudrate grsser 1 Prozent und damit zu hoch!"
.endif
 
.dseg

SerialRxZeichen:
	.byte 1
.ifdef SERIAL_HEX_INPUT
SerialHexInput:
	.byte 1
SerialHexInputNdx:
	.byte 1
.endif

.cseg

SerialInit:
	; Serielle Schnittstelle initialisieren
	push r17
	in r17,SREG
	cli
	push r16

	ldi r16,0
	sts SerialRxZeichen,r16

.ifdef SERIAL_HEX_INPUT
	ldi r16,-1 ; Keine HEX-Eingabe im Gange
	sts SerialHexInputNdx, r16
.endif

	;  Baudrate 9600
	;   UBRR auf 105 setzen
	;   16000000 Hz + (9600 baud * 8)
	;   -----------------------------
	;         (9600 baud * 16)
	ldi r16,HIGH(SERIAL_UBRR_VAL)
	out UBRRH,r16
	ldi r16,LOW(SERIAL_UBRR_VAL)
	out UBRRL,r16


	;  Empfnger und Sender enablen, RX und TX Complete Interrupts
	;   Das USART Control and Status Register UCSRB
	;    RXCIE/TXCIE - Receive/Transmit Interrupt enable
	;    UDRIE - Data Register Empty Interrupt enable
	;	 RXEN/TXEN - Receive/Transmit enable
	;    UCSZ2 - Character Size
	;    RXB8/TXB8 - Receive/Transmit 9th Data Bit
	;    Bit 7		6		5		4		3		2		1		0
	;       RXCIE	TXCIE	UDRIE	RXEN	TXEN	UCSZ2	RXB8	TXB8
	;		1       1       1 		1		1		0		0		0
	; ldi r16,(1<<RXEN)|(1<<TXEN)|(1<<RXCIE)|(1<<TXCIE)|(1<<UDRIE)
	; ldi r16,(1<<RXEN)|(1<<TXEN)|(1<<RXCIE)|(1<<UDRIE)
	ldi r16,(1<<RXEN)|(1<<TXEN)|(1<<RXCIE)
	; ldi r16,(1<<RXEN)|(1<<TXEN)|(1<<RXCIE)|(1<<TXCIE)
	; ldi r16,(1<<RXEN)|(1<<TXEN)
	out UCSRB,r16
	; Frame Format 8N1
	;   Das USART Control and Status Register UCSRC
	;   Bit 7		6		5		4		3		2		1		0
	;       URSEL	UMSEL	UPM1	UPM0	USBS	UCSZ1	UCSZ0	UCPOL
	; 		1		0		0		0		0		1		1		0
	ldi r16,(1<<URSEL)|(3<<UCSZ0)
	out UCSRC,r16

	pop r16
	out SREG,r17
	pop r17
	ret

SerialTXCharNow:
	; sbi UCSRB,TXEN
	; Parameter r16
	sbis UCSRA,UDRE
	rjmp SerialTXCharNow
	out UDR,r16
	ret

SerialRxChar:
	; sbi UCSRB,RXEN
	sbis UCSRA,RXC
	rjmp SerialRxChar
	in r16,UDR
	ret
/*
SregAusgebenMeldung:
	.db "SREG:",13,10,"ITHS NZC",13,10,0
SregAusgeben:
	push r16
	in r16,SREG
	push r16
	push ZL
	push ZH
	ldi ZL,LOW(SregAusgebenMeldung*2)
	ldi ZH,HIGH(SregAusgebenMeldung*2)
	rcall SerialStringAusgeben
	rcall BinaerAusgeben
	pop ZH
	pop ZL
	pop r16
	out SREG,r16
	pop r16
	ret
	
*/

.ifdef SERIAL_HEX_OUTPUT

SerialFromHexadezimal:
	; errechnet aus der HEX-Ziffer in r16 den Wert und legt ihn in r16 ab. (T-Flag, wenn Fehler)
	cpi   r16,'0'
	brlo  SerialFromHexadezimalFehler
	cpi   r16,'9'+1
	brlo  SerialFromHexadezimalDezZiffer
	andi  r16,0b11011111 ; 'a' => 'A'
	cpi   r16,'A'
	brlo  SerialFromHexadezimalFehler
	cpi   r16,'F'+1
	brlo  SerialFromHexadezimalBuchstabe
SerialFromHexadezimalFehler:
	set
	ret
SerialFromHexadezimalBuchstabe:
	subi   r16,'A'-('9'+1)
SerialFromHexadezimalDezZiffer:
	subi   r16,'0'
	clt
	ret

SerialHexadezimal:
	; legt die Hexadezimale Repsentation von r16 in r17 (low) und r18 (high) also in der Form "0x<r18><r17>" ab.
	push  r19
	push  r20
	mov   r17,r16
	mov   r18,r16
	ldi   r19,'0'
	ldi   r20,'A'-10
	andi  r17,0b00001111
	lsr   r18
	lsr   r18
	lsr   r18
	lsr   r18
	cpi    r17,10
	brge  SerialBuchstabeZiffer2
	add   r17,r19
	rjmp  SerialZiffer2Weiter
SerialBuchstabeZiffer2:
	add   r17,r20
SerialZiffer2Weiter:
	cpi    r18,10
	brge  SerialBuchstabeZiffer1
	add   r18,r19
	rjmp  SerialZiffer1Weiter
SerialBuchstabeZiffer1:
	add   r18,r20
SerialZiffer1Weiter:
	pop r20
	pop r19
	ret

SerialHexadezimalAusgeben:
	push r15
	in r15,SREG
	push r17
	push r18
	rcall SerialHexadezimal
SerialZiffer1AusgebenWarten:
	sbis UCSRA,UDRE
	rjmp SerialZiffer1AusgebenWarten
	out UDR,r18
SerialZiffer2AusgebenWarten:
	sbis UCSRA,UDRE
	rjmp SerialZiffer2AusgebenWarten
	out UDR,r17
	pop r18
	pop r17
	out SREG,r15
	pop r15
	ret

SerialByteAusgeben:
	; push  r15
	; in    r15,SREG
	push  r16
	rcall SerialHexadezimalAusgeben
	ldi   r16,' '
	rcall SerialTxCharNow
	; ldi r16,10
	; rcall SerialTxCharNow
	pop   r16
	; out   SREG,r15
	; pop   r15
	ret

.endif

.ifdef SERIAL_BIN
SerialBinaerAusgeben:
	push r15
	in r15,SREG
	push r16
	push r17
	push r18
	push r19
	ldi r19,8
	ldi r18,'0'
	clc
SerialBinaerAusgebenSchleife:
	ldi r17,0
	rol r16
	adc r17,r18
SerialBinaerAusgebenWarten:
	sbis UCSRA,UDRE
	rjmp SerialBinaerAusgebenWarten
	out UDR,r17
	dec r19
	brne SerialBinaerAusgebenSchleife
	/*ldi r16,13
	rcall SerialTxCharNow
	ldi r16,10
	rcall SerialTxCharNow*/
	pop r19
	pop r18
	pop r17
	pop r16
	out SREG,r15
	pop r15
	ret
.endif

SerialStringAusgeben:
	; push r15
	; in r15,SREG
	push r18
	push r30
	push r31
SerialStringAusgebenSchleife:
	lpm r18,Z+
	cpi r18,0
	breq SerialStringAusgebenEnde
SerialStringAusgebenWarten:
	sbis UCSRA,UDRE
	rjmp SerialStringAusgebenWarten
	out UDR,r18
	rjmp SerialStringAusgebenSchleife
SerialStringAusgebenEnde:
	pop r31
	pop r30
	pop r18
	; out SREG,r15
	; pop r15
	ret

.ifdef SERIAL_HEX_INPUT
SerialSetHexInput:
	push  r16
	ldi   r16,0
	sts   SerialHexInputNdx, r16
	pop   r16
	ret
.endif
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Interrupt-Routinen:

USART_RXC: ; USART RX Complete Handler
	push  r16	; r16 sichern
	in    r16,SREG
	push  r16	; SREG sichern
	in    r16,UDR	; RX-Datenregister lesen (Zeichen empfangen)
	sts   SerialRxZeichen,r16
.ifdef SERIAL_HEX_INPUT
	clt
	push  r17
	lds   r17, SerialHexInputNdx
	cpi   r17,-1
	breq  USART_RXC_Weiter
	; Hex-Ziffer?
	rcall SerialFromHexadezimal
	brts  USART_RXC_HexFehler
	cpi   r17,1
	brge  USART_RXC_HexFertig
	swap  r16 
	sts   SerialHexInput, r16
	inc   r17
	sts   SerialHexInputNdx, r17
	pop   r17
	rjmp  USART_RXC_Ende
USART_RXC_HexFehler:
	ldi   r17, -1
	sts   SerialHexInputNdx, r17
	rjmp  USART_RXC_Weiter
USART_RXC_HexFertig:
	ldi   r17, -1
	sts   SerialHexInputNdx, r17
	lds   r17, SerialHexInput
	or    r16, r17
USART_RXC_Weiter:
	pop   r17
.endif
	rcall SerialZeichenVerarbeiten
USART_RXC_Ende:
	pop   r16 	; SREG restaurieren
	out   SREG,r16
	pop   r16	; r16 restaurieren
	reti

USART_UDRE: ; UDR Empty Handler
	reti

USART_TXC: ; USART TX Complete Handler
	reti
